module HAMSI_INTERFACE(
	clk, rst_n,
	init, EOM, load, fetch, idata,                                /* input from Control FPGA */
	ack, odata,                                                   /* output for Control FPGA */
	busy, hash0, hash1, hash2, hash3, hash4, hash5, hash6, hash7, /* input from HAMSI_CORE */
	init_r, EOM_r, EN, idata_r                                    /* output for HAMSI_CORE */
);  

//{{{/* Input/Output */
/* From/For Control FPGA */
input         clk, rst_n;
input         init, EOM, load, fetch;
input  [15:0] idata;
output        ack;
output [15:0] odata;

/* From/For HAMSI_CORE */
input         busy;
input  [31:0] hash0, hash1, hash2, hash3, hash4, hash5, hash6, hash7;
output        init_r, EOM_r, EN;
output [31:0] idata_r;
//}}}

//{{{/* register and wire */
/* init/EOM register */
reg        init_r, EOM_r;

/* load/fetch register */
reg        load_r, fetch_r;

/* acknowledgement signal register */
reg        ack_r;

/* message value register */
reg [31:0] idata_r;
reg [15:0] odata_r;

/* data count */
reg [3:0]  data_count;

/* state machine */
reg [2:0]  state;
reg [2:0]  next_state;
//}}}

//{{{/* enable */
assign EN = ((state == 3'b010) && (data_count[0] == 1'b0))? 1'b1 : 1'b0; /* when hash execution state and input data_count is odd */
//}}}

//{{{/* init/EOM */
/* init */
always @(posedge clk or negedge rst_n) begin
	if (~rst_n) begin
		init_r <= 1'b0;
	end
	else begin
		if (init) begin
			init_r <= 1'b1;
		end
		else begin
			init_r <= 1'b0;
		end
	end
end

/* EOM */
always @(posedge clk or negedge rst_n) begin
	if (~rst_n) begin
		EOM_r <= 1'b0;
	end
	else begin
		if (EOM) begin
			if (state == 3'b000) begin /* EOM ɂāCconst ̒lvZrŕωȂ悤ɂD */
				EOM_r <= 1'b1;
			end
			else begin
				EOM_r <= EOM_r;
			end
		end
		else begin
			EOM_r <= EOM_r;
		end
	end
end
//}}}

//{{{/* load/fetch */
/* load */
always @(posedge clk or negedge rst_n) begin
	if (~rst_n) begin
		load_r <= 1'b0;
	end
	else begin
		if (load) begin
			load_r <= 1'b1;
		end
		else begin
			load_r <= 1'b0;
		end
	end
end

/* fetch */
always @(posedge clk or negedge rst_n) begin
	if (~rst_n) begin
		fetch_r <= 1'b0;
	end
	else begin
		if (fetch) begin
			fetch_r <= 1'b1;
		end
		else begin
			fetch_r <= 1'b0;
		end
	end
end
//}}}

//{{{/* ack */
assign ack = ack_r;
always @(posedge clk or negedge rst_n)begin
	if (~rst_n) begin
		ack_r <= 1'b0;
	end
	else begin
		if (state == 3'b001 || state == 3'b100) begin /* if load state (3'b001) or fetch state (3'b100) */
			ack_r <= 1'b1;
		end
		else begin
			ack_r <= 1'b0;
		end
	end
end
//}}}

//{{{/* output data */
assign odata = odata_r;
always @(posedge clk or negedge rst_n)begin
	if (~rst_n) begin
		odata_r <= 16'h0;
	end
	else begin
		if (state == 3'b011) begin
			if (data_count == 4'd0) begin
				odata_r <= hash0[31:16];
			end
			else if (data_count == 4'd1) begin
				odata_r <= hash0[15: 0];
			end
			else if (data_count == 4'd2) begin
				odata_r <= hash1[31:16];
			end
			else if (data_count == 4'd3) begin
				odata_r <= hash1[15: 0];
			end
			else if (data_count == 4'd4) begin
				odata_r <= hash2[31:16];
			end
			else if (data_count == 4'd5) begin
				odata_r <= hash2[15: 0];
			end
			else if (data_count == 4'd6) begin
				odata_r <= hash3[31:16];
			end
			else if (data_count == 4'd7) begin
				odata_r <= hash3[15: 0];
			end
			else if (data_count == 4'd8) begin
				odata_r <= hash4[31:16];
			end
			else if (data_count == 4'd9) begin
				odata_r <= hash4[15: 0];
			end
			else if (data_count == 4'd10) begin
				odata_r <= hash5[31:16];
			end
			else if (data_count == 4'd11) begin
				odata_r <= hash5[15: 0];
			end
			else if (data_count == 4'd12) begin
				odata_r <= hash6[31:16];
			end
			else if (data_count == 4'd13) begin
				odata_r <= hash6[15: 0];
			end
			else if (data_count == 4'd14) begin
				odata_r <= hash7[31:16];
			end
			else if (data_count == 4'd15) begin
				odata_r <= hash7[15: 0];
			end
		end
		else begin
			odata_r <= odata_r;
		end
	end
end
//}}}

//{{{/* data counter */
always @(posedge clk or negedge rst_n) begin
	if (~rst_n) begin
		data_count <= 4'h0;
	end
	else begin
		if (state == 3'b001) begin      /* if load state */
			if (data_count == 4'd1) begin
				if (~busy) begin
					data_count <= 4'h0;
				end
				else begin
					data_count <= data_count;
				end
			end
			else begin
				data_count <= data_count + 1'd1;
			end
		end
		else if (state == 3'b011) begin /* if fetch state */
			if (data_count == 4'd15) begin
				data_count <= 4'h0;
			end
			else begin
				data_count <= data_count + 1'd1;
			end
		end
		else begin
			data_count <= data_count;
		end
	end
end
//}}}

//{{{/* input data */
always @(posedge clk or negedge rst_n) begin
	if (~rst_n) begin
		idata_r <= 32'h0;
	end
	else begin
		if (state == 3'b001) begin /* if load state */
			idata_r <= {idata_r[15:0],idata};
		end
		else begin
			idata_r <= idata_r;
		end
	end
end
//}}}

//{{{/* Finite State Machine */
always @(posedge clk or negedge rst_n) begin
	if (~rst_n) begin
		state <= 3'b000;
	end
	else begin
		state <= next_state;
	end
end

always @(load_r or fetch_r or EN or busy or state) begin
	case (state)
	3'b000 : begin /* neutral state */
		if (load_r) begin
			next_state <= 3'b001;
		end
		else if (fetch_r) begin
			next_state <= 3'b011;
		end
		else begin
			next_state <= 3'b000;
		end
	end
	3'b001 : begin /* load state */
		next_state <= 3'b010;
	end
	3'b010 : begin /* execution state */
		if (busy) begin
			next_state <= 3'b010;
		end
		else begin
			next_state <= 3'b000;
		end
	end
	3'b011 : begin /* fetch state */
		next_state <= 3'b100;
	end
	3'b100 : begin /* output state */
		next_state <= 3'b000;
	end
	default : begin
		next_state <= 3'b000;
	end
        endcase
end

endmodule
//}}}

// // vim:fdm=marker
